#include "MySQLDS.h"
#include <stdio.h>
#include <stack>
#include <vector>
#include <list>
#include <ctype.h>
#include <utility>
//#if MACHINE_LINUX
#include <memory>
//#endif

const char * kRowCountVar = "__updated_count__";

#ifdef WIN32
#define strcasecmp stricmp
#endif

#define DEBUG_SQL

MYSQL * local_doConnect( lasso_request_t token );

void local_addReturnFields( lasso_request_t token, STRING_TYPE & to );
void local_addSearchCriteria( lasso_request_t token, STRING_TYPE & to, LP_TypeDesc defaultOp, bool isFindAll );
void local_addUpdateCriteria( lasso_request_t token, STRING_TYPE & to );
void local_addSortOrder( lasso_request_t token, STRING_TYPE & to );
void local_addSearchValue( STRING_TYPE & to, const char * value, unsigned int size, LP_TypeDesc op );
bool local_addPrimaryKeySearch( lasso_request_t token, STRING_TYPE & to );
void local_addColumnInfo( lasso_request_t token, MYSQL_RES * res, MYSQL * mysql, const char * db, const char * table );
void local_addOperator( STRING_TYPE & to, LP_TypeDesc op, const char * value, unsigned int size );
void local_addInsertCriteria( lasso_request_t token, STRING_TYPE & query );

void registerLassoModule()
{
	lasso_registerDSModule( "MySQLDS", local_mysqlds_func, 0 );
	lasso_log(LOG_LEVEL_CRITICAL, "MySQLConnector using MySQL client: %s",  mysql_get_client_info());
}

inline osError CheckError(lasso_request_t token, MYSQL * m)
{
	if ( mysql_errno(m))
	{
		lasso_log(LOG_LEVEL_WARNING, "MySQLDS: error from mysql: %d, %s",  mysql_errno(m),  mysql_error(m));
		lasso_setResultMessage(token,  mysql_error(m));
		return  mysql_errno(m);
	}
	return osErrNoErr;
}

inline void AddColumnName(const char * in, STRING_TYPE & out)
{
	if (in)
	{
		int x = 0;
		out.APND("`");
		if (in[x] == '`')
			++x;
		while(in[x])
		{
			if (in[x] == '`' || in[x] == '#' || (in[x] == '-' && in[x+1] == '-') || (in[x] == '/' && in[x] == '*'))
				break;
			out.APND(in+x, 1);
			++x;
		}
		out.APND("`");
	}
}

MYSQL * local_doConnect( lasso_request_t token )
{
	// updated to work with LP7 datasource connection caching
	lasso_dsconnection_t test = NULL;
	if (lasso_getDSConnection(token, &test)==osErrNoErr && test != NULL)
	{		
//		lasso_log(LOG_LEVEL_DETAIL, "MySQLDS using cached connection %d\n", (int)test);
		return reinterpret_cast<MYSQL*>(test);
	}
	auto_lasso_value_t host, auth;
	lasso_getDataHost(token, &host, &auth);

	MYSQL * worked = NULL;
	
	worked =  mysql_init(NULL);

	if ( mysql_real_connect(worked, host.name, auth.name, auth.data, NULL, atoi(host.data), "/tmp/mysql.sock", CLIENT_FOUND_ROWS) == NULL )
	{
		lasso_log(LOG_LEVEL_CRITICAL, "MySQLDS failed connecting to MySQL. errno=%d \"%s\"",  mysql_errno(worked),  mysql_error(worked));
		 mysql_close(worked);
		worked = NULL;
	}
	else
	{
//		lasso_log(LOG_LEVEL_DETAIL, "MySQLDS caching new connection %d\n", (int)worked);
		lasso_setDSConnection(token, reinterpret_cast<lasso_dsconnection_t>(worked));
	}
	return worked;
}

osError local_mysqlds_func( lasso_request_t token, datasource_action_t action, const auto_lasso_value_t * param )
{
	osError err = osErrNoErr;
	switch( action )
	{
		case datasourceInit:
			// nothing!
			break;
		case datasourceTerm:
			// nothing!
			break;
		case datasourceCloseConnection: // connections only get closed through here
			{
				lasso_dsconnection_t conn = NULL;
				if ( lasso_getDSConnection(token, &conn) == osErrNoErr && conn )
				{
//					lasso_log(LOG_LEVEL_DETAIL, "MySQLDS closing cached connection %d\n", (int)conn);
					 mysql_close(reinterpret_cast<MYSQL*>(conn));
				}
			}
			break;
		case datasourceTickle:
			{
				MYSQL * conn = local_doConnect(token);
				auto_lasso_value_t db;

				lasso_getDataSourceName(token, &db, NULL, NULL);
				 mysql_select_db(conn, db.name);
			}
			break;
		case datasourceNames:
			{
				MYSQL * mysql = local_doConnect(token);
				if ( mysql == NULL )
					return osErrConnectionInvalid;
				
				MYSQL_RES * res = NULL;
				MYSQL_ROW row;
				
				if (  mysql_query(mysql, "show databases") == 0 )
				{
					res =  mysql_store_result(mysql);
					if ( res != NULL )
					{
						while( (row =  mysql_fetch_row(res)) != NULL )
						{
							lasso_addDataSourceResult( token, row[0] );
						}
						 mysql_free_result(res);
					}
				}
//				mysql_close(mysql);				
			}
			break;
		case datasourceTableNames:
			{
				MYSQL_RES * res = NULL;

				auto_lasso_value_t db;

				lasso_getDataSourceName(token, &db, NULL, NULL);
				
				MYSQL * mysql = local_doConnect(token);
				if ( mysql == NULL )
					return osErrConnectionInvalid;

				if (  mysql_select_db(mysql, db.data) == 0 )
				{
					res =  mysql_list_tables(mysql, NULL);
					if ( res != NULL )
					{
						MYSQL_ROW row;
						while( (row =  mysql_fetch_row(res)) != NULL )
						{
							lasso_addDataSourceResult(token, row[0]);
						}
						 mysql_free_result(res);
					}
				}
				err = CheckError(token, mysql);
			}
			break;
		case datasourceSearch:
		case datasourceFindAll:
			{
				auto_lasso_value_t dbName, tableName, test;
				
				STRING_TYPE query;
			
				lasso_getDataSourceName( token, &dbName, NULL, NULL );
				lasso_getTableName( token, &tableName );
					
				MYSQL * mysql = local_doConnect(token);
				if ( mysql == NULL )
					return osErrConnectionInvalid;

				/* SELECT (<result_fields>) FROM <databasename>.<table_name> WHERE <search_fields> ORDER BY <sort_fields> */
				query.APND("SELECT ");
				
				if ( lasso_findInputColumn( token, "-distinct", &test ) == osErrNoErr )
				{
					query.APND("DISTINCT ");
				}

				local_addReturnFields( token, query );
				
				query.APND(" FROM ").APND( dbName.data ).APND(".").APND( tableName.data );

				int tmp;
				lasso_getRowID(token, &tmp);
				if ( tmp != kLassoAny )
				{
					local_addSearchCriteria( token, query, lpTypeNull, action == datasourceFindAll );
	#ifdef DEBUG_SQL
					lasso_log(LOG_LEVEL_DETAIL, "MySQLDS made query: %s", query.CHAR_PTR());
	#endif				
					if (  mysql_query(mysql, query.CHAR_PTR()) == 0 )
					{
						MYSQL_ROW row;
						MYSQL_RES * res =  mysql_store_result(mysql);
							
						if ( res != NULL )
						{
							int num =  mysql_num_fields(res), skip=0, max=0, found;
							// we have to set the column info
							local_addColumnInfo( token, res, mysql, dbName.data, tableName.data );
					
							// if the query used LIMIT then don't do skip and max. just get them all
							if ( lasso_findInputColumn( token, "-uselimit", &test ) == osErrNoErr )
							{
								max = -1;
							}
							else
							{
								lasso_getSkipRows( token, &skip );
								lasso_getMaxRows( token, &max );
							}

							found = (int) mysql_num_rows(res);

							lasso_setNumRowsFound( token, found );

							if ( max > -1 && skip > 0 )
								 mysql_data_seek( res, skip );

							while( (row =  mysql_fetch_row(res)) != NULL && (max == -1 || max--) )
							{
								unsigned long * lengths =  mysql_fetch_lengths(res);
								lasso_addResultRow( token, (const char**)row, (unsigned int*)lengths, num );
							}
							 mysql_free_result(res);
						}
					}
				}
				else
				{
					query.APND(" ORDER BY RAND() LIMIT 1");
	#ifdef DEBUG_SQL
					lasso_log(LOG_LEVEL_DETAIL, "MySQLDS made query: %s", query.CHAR_PTR());
	#endif
					MYSQL_ROW row;
					MYSQL_RES * res = NULL;
			
					int result =  mysql_query(mysql, query.CHAR_PTR());
					if ( result == 0 )
					{
						res =  mysql_store_result(mysql);
					}	 

					if ( res != NULL )
					{
						int num =  mysql_num_fields(res);
						// we have to set the column info
						local_addColumnInfo( token, res, mysql, dbName.data, tableName.data );
						int found = (int) mysql_num_rows(res);

						lasso_setNumRowsFound( token, found );

						row =  mysql_fetch_row(res);
						if ( row != NULL )
						{
							unsigned long * lengths =  mysql_fetch_lengths(res);
							lasso_addResultRow( token, (const char**)row, (unsigned int*)lengths, num );
						}
						 mysql_free_result(res);
					}					
				}				
				err = CheckError(token, mysql);
			}
			break;
		case datasourceAdd:
			 {				 
				auto_lasso_value_t dbName, tableName;

				lasso_getDataSourceName( token, &dbName, NULL, NULL );
				lasso_getTableName( token, &tableName );
				
				MYSQL * mysql = local_doConnect(token);
				if ( mysql == NULL )
					return osErrConnectionInvalid;

				STRING_TYPE query;

				query.APND("INSERT INTO ").APND(dbName.data).APND(".").APND(tableName.data);

				local_addInsertCriteria(token, query);
#ifdef DEBUG_SQL
				lasso_log(LOG_LEVEL_DETAIL, "MySQLDS made query: %s", query.CHAR_PTR());
#endif
				if ( mysql_query( mysql, query.CHAR_PTR() ) == 0)
				{
					if (  mysql_affected_rows(mysql) == 0 )
					{
						err = osErrWebAddError;
						lasso_setVariable( token, kRowCountVar, "0", 0 );
					}
					else
					{
						char rowcount[16] = "";      
						sprintf( rowcount, "%lu", (unsigned long) mysql_affected_rows(mysql) );
						lasso_setVariable( token, kRowCountVar, rowcount, 0 );
					}
					if( err == osErrNoErr )
					{
						lasso_setRowID( token, (unsigned long)  mysql_insert_id( mysql ) );
						int c;
						auto_lasso_value_t key;
						lasso_getMaxRows(token, &c);
						if ( c > 0 && lasso_getPrimaryKeyColumn( token, &key ) == osErrNoErr && key.nameSize > 0 )
						{
							query = "SELECT ";
							local_addReturnFields(token, query);
							query.APND(" FROM ").APND(dbName.data).APND(".").APND(tableName.data).APND(" WHERE ").APND(key.name).APND(" = LAST_INSERT_ID()");
							int result =  mysql_query(mysql, query.CHAR_PTR());
							if ( result == 0 )
							{
								MYSQL_RES * res =  mysql_store_result(mysql);
								int num =  mysql_num_fields(res);
								local_addColumnInfo( token, res, mysql, dbName.data, tableName.data );
								MYSQL_ROW row =  mysql_fetch_row(res);
								if ( row != NULL )
								{
									unsigned long * lengths =  mysql_fetch_lengths(res);
									lasso_addResultRow( token, (const char**)row, (unsigned int*)lengths, num );
									 mysql_free_result(res);
								}
							}
						}
					}
				}
				err = CheckError(token, mysql);
			}
			break;
		case datasourceUpdate:
			 {
				auto_lasso_value_t db, table;
				lasso_getDataSourceName(token, &db, NULL, NULL);
				lasso_getTableName(token, &table);
				
				MYSQL * mysql = local_doConnect(token);
				if ( mysql == NULL )
					return osErrConnectionInvalid;

				STRING_TYPE query;

				query.APND("UPDATE ").APND(db.data).APND(".").APND(table.data);

				local_addUpdateCriteria( token, query );
				if ( !local_addPrimaryKeySearch( token, query ) )
				{
					return osErrWebUpdateError;
				}
#ifdef DEBUG_SQL
				lasso_log(LOG_LEVEL_DETAIL, "MySQLDS made query: %s", query.CHAR_PTR());
#endif
				if ( mysql_query(mysql, query.CHAR_PTR()) == 0)
				{				
					if (  mysql_affected_rows(mysql) > 0 )
					{
						lasso_setVariable( token, kRowCountVar, "0", 0 );
						int c;
						lasso_getMaxRows(token, &c);
						if ( c > 0 )
						{
							query = "SELECT ";
							local_addReturnFields(token, query);
							query.APND(" FROM ").APND(db.data).APND(".").APND(table.data);
							local_addPrimaryKeySearch( token, query );

							int result =  mysql_query(mysql, query.CHAR_PTR());
							if ( result == 0 )
							{
								MYSQL_RES * res =  mysql_store_result(mysql);
								int num =  mysql_num_fields(res);
								local_addColumnInfo( token, res, mysql, db.data, table.data );
								MYSQL_ROW row =  mysql_fetch_row(res);
								if ( row != NULL )
								{
									unsigned long * lengths =  mysql_fetch_lengths(res);
									lasso_addResultRow( token, (const char**)row, (unsigned int*)lengths, num );
									 mysql_free_result(res);
								}
							}
						}
					}
					else
					{
						char rowcount[16] = "";      
						sprintf( rowcount, "%lu", (unsigned long)  mysql_affected_rows(mysql) );
						lasso_setVariable( token, kRowCountVar, rowcount, 0 );
					}
				}
				err = CheckError(token, mysql);
			}
			break;
		case datasourceDelete:
			 {
				auto_lasso_value_t db, table;

				lasso_getDataSourceName( token, &db, NULL, NULL );
				lasso_getTableName( token, &table );
					
				MYSQL * mysql = local_doConnect(token);
				if ( mysql == NULL )
					return osErrConnectionInvalid;

				STRING_TYPE query;

				query.APND("DELETE FROM ").APND(db.data).APND(".").APND(table.data);
				if ( !local_addPrimaryKeySearch( token, query ) )
				{
					return osErrWebDeleteError;
				}
#ifdef DEBUG_SQL
				lasso_log(LOG_LEVEL_DETAIL, "MySQLDS made query: %s", query.CHAR_PTR());
#endif
				if ( mysql_query(mysql, query.CHAR_PTR()) == 0)
				{
					if (  mysql_affected_rows(mysql) == 0 )
					{
						err = osErrWebDeleteError;
						lasso_setVariable( token, kRowCountVar, "0", 0 );
					}
					else
					{
						char rowcount[16] = "";      
						sprintf( rowcount, "%lu", (unsigned long) mysql_affected_rows(mysql) );
						lasso_setVariable( token, kRowCountVar, rowcount, 0 );
					}			    
				}
				err = CheckError(token, mysql);
			}
			break;
		case datasourceInfo:
			 {
				auto_lasso_value_t dbName;
				auto_lasso_value_t tableName;

				lasso_getDataSourceName( token, &dbName, NULL, NULL );
				lasso_getTableName( token, &tableName );
				if ( tableName.data[0] != '\0' )
				{
					MYSQL * mysql = local_doConnect(token);
					if ( mysql == NULL )
						return osErrConnectionInvalid;
									
					char query[256] = "";
					/* no result set but we get the column info */
					sprintf(query, "SELECT * FROM %s.%s WHERE 1=0", dbName.data, tableName.data);

					if (  mysql_query(mysql, query) == 0 )
					{
						MYSQL_RES * res =  mysql_store_result(mysql);
						
						local_addColumnInfo( token, res, mysql, dbName.data, tableName.data );
						
						 mysql_free_result(res);
					}
					err = CheckError(token, mysql);
				}
			}
			break;
		case datasourceExecSQL:
			if ( param != NULL && param->data != NULL )
			{				
				MYSQL * mysql = local_doConnect(token);
//				printf("datasourceExecSQL: %s\n",  mysql_get_server_info(mysql));
				if ( mysql == NULL )
					return osErrConnectionInvalid;
				
				MYSQL_RES * res = NULL;
				auto_lasso_value_t dbName;
				lasso_getDataSourceName( token, &dbName, NULL, NULL );
				
				 mysql_select_db( mysql, dbName.data );	
				
				int max = param->dataSize, start=-1, finish=-1;
				while( err == osErrNoErr && lasso_sqlFindStat(param->data, &start, &finish, max) )
				{
					if ( res )
					{
						 mysql_free_result(res);
						res = NULL;
					}
					if ( mysql_real_query( mysql, param->data + start, finish-start ) == 0)
						res =  mysql_store_result(mysql);
					else
						err =  mysql_errno(mysql);
				}
				if ( err == osErrNoErr )
				{
					lasso_setRowID( token, (unsigned long)  mysql_insert_id( mysql ) );

					char rowcount[16] = "";      
					sprintf( rowcount, "%lu", (unsigned long) mysql_affected_rows(mysql) );
					lasso_setVariable( token, kRowCountVar, rowcount, 0 );

					if ( res != NULL )
					{
						local_addColumnInfo(token, res, mysql, "", "");

						int found = (int) mysql_num_rows(res);
						int num =  mysql_num_fields(res);
						int skip, max;
						MYSQL_ROW row;

						lasso_getSkipRows( token, &skip );
						lasso_getMaxRows( token, &max );
						lasso_setNumRowsFound( token, found );
					
						if ( skip > 0 )
							 mysql_data_seek( res, skip );

						while( (row =  mysql_fetch_row(res)) != NULL && max--)
						{
							unsigned long * lengths =  mysql_fetch_lengths(res);
							lasso_addResultRow( token, (const char**)row, (unsigned int*)lengths, num );
						}
						 mysql_free_result(res);
					}
				}
				err = CheckError(token, mysql);
			}
			break;
		default:
			break;

	}
	return err;
}

struct string_ptr_pair_t
{
	string_ptr_pair_t( const STRING_TYPE & str, const std::auto_ptr<STRING_TYPE> & ptr)
	{
		first.CSET(str);
		second = const_cast<std::auto_ptr<STRING_TYPE>&>(ptr);
	}
	string_ptr_pair_t( const string_ptr_pair_t & pr )
	{
		first.CSET(pr.first);
		second = const_cast<string_ptr_pair_t&>(pr).second;
	}
	string_ptr_pair_t(){}

	STRING_TYPE first;
	std::auto_ptr<STRING_TYPE> second;
};
typedef std::list<string_ptr_pair_t> consolidated_input_t;

inline void AddOrCat( const char * name, const char * value, int valueLen, consolidated_input_t * lst )
{
	for ( consolidated_input_t::iterator it = lst->begin(), end = lst->end(); it != end; ++it )
	{
		if ( (*it).first.CEQUALS(name) ) // it's in there
		{
			if ( value != NULL )
			{
				if ( (*it).second.get() == NULL )
					(*it).second.reset(new STRING_TYPE);
				else
					(*it).second->APND(",");
				(*it).second->APND(value, valueLen);
			}
			return;
		}
	}
	// add it anew
	lst->push_back(string_ptr_pair_t(STRING_TYPE(name), std::auto_ptr<STRING_TYPE>(NULL)));
	if ( value != NULL )
		lst->back().second.reset(new STRING_TYPE(value, valueLen));
}

inline void
ConsolidateInput( lasso_request_t token, consolidated_input_t * lst )
{
	int num,i;

	lasso_getInputColumnCount( token, &num );	
	for ( i = 0; i < num; ++i )
	{
		auto_lasso_value_t val;
		if ( lasso_getInputColumn( token, i, &val ) == osErrNoErr )
		{
			if (val.name && val.name[0])
				AddOrCat( val.name, val.data, val.dataSize, lst);
		}
	}
}

void local_addInsertCriteria( lasso_request_t token, STRING_TYPE & query )
{
	int num, i;
	consolidated_input_t lst;
	ConsolidateInput(token, &lst);

	num = lst.size();
	if ( num > 0 )
	{
		query.APND(" (");
		i = 0;
		{for ( consolidated_input_t::iterator it = lst.begin(), end = lst.end(); it != end; ++it )
		{
			AddColumnName((*it).first.CHAR_PTR(), query);
			if ( i + 1 < num )
				query.APND(",");
			++i;
		}}
		query.APND(") VALUES (");
		i = 0;
		{for ( consolidated_input_t::iterator it = lst.begin(), end = lst.end(); it != end; ++it )
		{
			const char * str = (*it).second.get() ? (*it).second->CHAR_PTR() : NULL;
			int len = (*it).second.get() ? (*it).second->CLENGTH() : 0;
			local_addSearchValue( query, str, len, lpTypeNull);
			
			if ( i + 1 < num )
				query.APND(",");
			++i;
		}}
		query.APND(")");
	}	
}

void local_addUpdateCriteria( lasso_request_t token, STRING_TYPE & to )
{
	consolidated_input_t lst;
	bool addComma = false;

	ConsolidateInput( token, &lst );

	if ( lst.size() > 0 )
	{
		to.APND(" SET ");
		for ( consolidated_input_t::iterator it = lst.begin(), end = lst.end(); it != end; ++it )
		{
			if ( addComma )
				to.APND(",");
			else
				addComma = true;

			AddColumnName((*it).first.CHAR_PTR(), to);
			to.APND(" = ");
			const char * str = (*it).second.get() ? (*it).second->CHAR_PTR() : NULL;
			int len = (*it).second.get() ? (*it).second->CLENGTH() : 0;
			local_addSearchValue( to, str, len, lpTypeNull );		
		}
	}
}

bool local_addPrimaryKeySearch( lasso_request_t token, STRING_TYPE & to )
{
	auto_lasso_value_t key;
	
	if ( lasso_getPrimaryKeyColumn( token, &key ) == osErrNoErr )
	{
		if ( key.name != NULL && strlen(key.name) > 0 )
		{
			to.APND(" WHERE ");
			AddColumnName(key.name, to);
			to.APND(" = ");
			if ( key.data != NULL && strlen(key.data) > 0 )
			{
				local_addSearchValue( to, key.data, key.dataSize, lpTypeNull );
				return true;
			}
		}
	}
	return false;
}
/*
+-------+--------------------------------+------+-----+---------+-------+
| Field | Type                           | Null | Key | Default | Extra |
+-------+--------------------------------+------+-----+---------+-------+
| Title | enum('Mr.','Mrs.','Ms.','Dr.') | YES  | MUL | NULL    |       |
+-------+--------------------------------+------+-----+---------+-------+
*/
void grabValueList( MYSQL * mysql, const char * col, const char * db, const char * table, char *** vals, int * count )
{
	STRING_TYPE query;
	query.APND("SHOW COLUMNS FROM ").APND(table).APND(" FROM ").APND(db).APND(" LIKE '").APND(col).APND("'");

	int result =  mysql_query(mysql, query.CHAR_PTR());
	if ( result == 0 )
	{
		MYSQL_RES * res =  mysql_store_result(mysql);
		if ( res )
		{
			int num =  mysql_num_fields(res);
			MYSQL_ROW row =  mysql_fetch_row(res);
			if ( row != NULL && num >= 2 )
			{
				const char * def = row ? row[1] : NULL;

				if ( def && (strncmp("enum", def, 4) == 0 || strncmp("set", def, 3) == 0) )
				{
					if ( def[0] == 'e' )
						def += 5;
					else
						def += 4;

					*count = 1;
					for ( int tmp = 0; def[tmp] != '\0'; ++tmp )
					{
						if ( def[tmp] == ',' )
							++*count;
					}

					*vals = new char*[*count];

					for ( int x = 0; x < *count; ++x )
					{
						if ( *def == '\'' )
							++def;
						STRING_TYPE str;
						while( *def != ',' && *def != '\'' )
						{
							str.APND(def, 0, 1);
							++def;
						}
						if ( *def == '\'' )
							def += 2;
						else
							++def;

						(*vals)[x] = new char[str.CLENGTH()+1];
						strcpy((*vals)[x], str.CHAR_PTR());
					}
				}
			}
			 mysql_free_result(res);
		}
	}
}

void local_addColumnInfo( lasso_request_t token, MYSQL_RES * res, MYSQL * mysql, const char * db, const char * table )
{
	int numFields =  mysql_num_fields(res), i;
	MYSQL_FIELD * field =  mysql_fetch_fields(res);	

	for( i = 0; i < numFields; ++i )
	{
		LP_TypeDesc type;
		switch( field[i].type )
		{
			case FIELD_TYPE_TINY:
			case FIELD_TYPE_SHORT:
			case FIELD_TYPE_LONG:
			case FIELD_TYPE_INT24:
			case FIELD_TYPE_LONGLONG:
				type = lpTypeInteger;
				break;
			case FIELD_TYPE_DECIMAL:
			case FIELD_TYPE_FLOAT:
			case FIELD_TYPE_DOUBLE:
				type = lpTypeDecimal;
				break;
			case FIELD_TYPE_TIMESTAMP:
			case FIELD_TYPE_DATE:
			case FIELD_TYPE_TIME:
			case FIELD_TYPE_DATETIME:
			case FIELD_TYPE_YEAR:
				type = lpTypeDateTime; // ???
				break;
			case FIELD_TYPE_TINY_BLOB:
			case FIELD_TYPE_MEDIUM_BLOB:
			case FIELD_TYPE_LONG_BLOB:
			case FIELD_TYPE_BLOB:
				if ( (field[i].flags & BINARY_FLAG) != 0)
				{
					type = lpTypeBytes;
					break;
				}
			default:
				type = lpTypeString;
				break;
		}		
		if ( (db != NULL && table != NULL && db[0] != '\0' && table[0] != '\0') && field[i].name && ((field[i].flags & ENUM_FLAG) != 0 || (field[i].flags & SET_FLAG) != 0) )
		{
			char ** vals = NULL;
			int count = 0;

			grabValueList(mysql, field[i].name, db, table, &vals, &count);
			lasso_addColumnInfo2( token, field[i].name, IS_NOT_NULL(field[i].flags) ? 0 : 1, type, kProtectionNone, (const char**)vals, count );

			for ( int o = 0; o < count; ++o )
				delete [] vals[o];
			delete [] vals;
		}
		else
			lasso_addColumnInfo( token, field[i].name, IS_NOT_NULL(field[i].flags) ? 0 : 1, type, kProtectionNone );
	}
}

void local_addReturnFields( lasso_request_t token, STRING_TYPE & to )
{
	int num, i;
	lasso_getReturnColumnCount( token, &num );

	if ( num == 0 )
	{
		to.APND("*");
	}
	else
	{
		for ( i = 0; i < num; ++i )
		{
			auto_lasso_value_t val;
			if ( lasso_getReturnColumn( token, i, &val ) == osErrNoErr )
			{
				if ( i > 0 )
					to.APND(",");
				AddColumnName(val.data, to);
			}
			else
				break;
		}
	}
}

#define ISOPBEGIN(X) (strcasecmp(X, kOpBegin) == 0 || strcasecmp(X, kOperatorBegin) == 0)
#define ISOPEND(X) (strcasecmp(X, kOpEnd) == 0 || strcasecmp(X, kOperatorEnd) == 0)

void local_addSearchCriteria( lasso_request_t token, STRING_TYPE & to, LP_TypeDesc defaultOp, bool isFindAll )
{
	int num, i;
	bool useLimit = false;
	auto_lasso_value_t val;
	LP_TypeDesc op;

	// if a primary key column was supplied then just use it for the search
	lasso_getPrimaryKeyColumn( token, &val );
	if ( !isFindAll && val.name != NULL && strlen(val.name) > 0 && val.data != NULL && strlen(val.data) > 0)
	{
		to.APND(" WHERE ");
		AddColumnName(val.name, to);
		
		char * encoded = new char[val.dataSize*2+1];
		mysql_escape_string(encoded, val.data, val.dataSize);
		to.APND(" = '").APND(encoded).APND("'");
		delete [] encoded;
	}
	else
	{
		lasso_getInputColumnCount( token, &num );

		if ( num > 0 )
		{		
			std::vector<auto_lasso_value_t> validValues;
			
			lasso_getLogicalOp( token, &op );
			auto_lasso_value_t tmpVal;
			tmpVal.name = "opbegin";

			if ( op == kLassoOR )
			{
				tmpVal.data = "or";
			}
			else
			{
				tmpVal.data = "and";
			}
			if (!isFindAll)
				validValues.push_back(tmpVal);
			
			for ( i = 0; i < num; ++i )
			{
				if ( lasso_getInputColumn( token, i, &val ) != osErrNoErr )
					return;
				if (val.name == NULL)
					continue;
				if ( ISOPEND(val.name) )
					val.data = "something";
				else if ( strcasecmp(val.name, "-uselimit") == 0 )
				{
					useLimit = true;
					continue;
				}
				else if ( val.name[0] == '-' && !ISOPBEGIN(val.name) )
				{
					continue;
				}
				// allow blank values
				if ( !isFindAll && ((val.data != NULL && strlen(val.data) > 0) || val.type == kLassoEquals || val.type == kLassoNOT) )
				{
					if ( i != 0 && ISOPEND(val.name) && ISOPBEGIN(validValues.back().name) )
					{
						validValues.pop_back();
					}
					else
						validValues.push_back(val);
				}
			}

			num = validValues.size();
			if ( num > 1 )
			{
				to.APND(" WHERE ");

				std::vector<LP_TypeDesc> opStack;

				for ( i = 0; i < num; ++i )
				{
					auto_lasso_value_t curr = validValues[i];
					if ( ISOPBEGIN(curr.name) )
					{
						if ( strcasecmp(curr.data, "not") == 0 )
						{
							to.APND("NOT (");
							opStack.push_back(kLassoNOT);
						}
						else
						{
							if ( strcasecmp(curr.data, "and") == 0 )
								opStack.push_back(kLassoAND);
							else if ( strcasecmp(curr.data, "or") == 0 )
								opStack.push_back(kLassoOR);
							else
								opStack.push_back(op);

							to.APND("(");
						}
						continue;
					}
					else if ( ISOPEND(curr.name) )
					{
						to.APND(")");
						opStack.pop_back();
					}
					else
					{
						if ( curr.type == kLassoInFullText )
						{
							to.APND("MATCH(");
							AddColumnName(curr.name, to);
							to.APND(")");
						}
						else
						{
							AddColumnName(curr.name, to);
						}
						
						if ( defaultOp != lpTypeNull ) // no default operator
						{
							curr.type = defaultOp;
						}

						local_addOperator( to, curr.type, curr.data, curr.dataSize );
						local_addSearchValue( to, curr.data, curr.dataSize, curr.type );
					}
					if ( (i + 1) < num && (!ISOPEND(validValues[i+1].name)) )
					{
						int num = opStack.size()-1;
						LP_TypeDesc lastOp = opStack[num--];
			
						while( num >= 0 && lastOp == kLassoNOT )
						{
							lastOp = opStack[num--];
						}

						if ( lastOp != kLassoAND )
							to.APND(" OR ");
						else
							to.APND(" AND ");
					}			
				}
				while ( opStack.size() > 0 )
				{
					to.APND(")");
					opStack.pop_back();
				}
			}
		}
	}
	local_addSortOrder( token, to );
	if ( useLimit )
	{
		int max, skip;
		lasso_getSkipRows( token, &skip );
		lasso_getMaxRows( token, &max );

		if ( max == -1 )
			max = 2147483647;
		char tmp[256];
		sprintf(tmp, " LIMIT %d,%d", skip, max);
		to.APND(tmp);
	}
}

void local_addSortOrder( lasso_request_t token, STRING_TYPE & to )
{
	int num, i;
	std::vector<auto_lasso_value_t> sorts;
	auto_lasso_value_t val;

	lasso_getSortColumnCount( token, &num );
	for ( i = 0; i < num; ++i )
	{
		if ( lasso_getSortColumn( token, i, &val ) == osErrNoErr )
		{
			if ( val.data != NULL && strlen(val.data) > 0 )
				sorts.push_back(val);
		}
		else
			break;
	}
	num = sorts.size();
	if ( num > 0 )
	{
		to.APND(" ORDER BY ");
		for ( i = 0; i < num; ++i )
		{
			AddColumnName(sorts[i].data, to);
			if ( sorts[i].type == orderDescending )
				to.APND(" DESC ");
			if ( i + 1 < num )
				to.APND(",");
		}
	}
	auto_lasso_value_t test;
	if ( lasso_findInputColumn(token, "-sortrandom", &test) == osErrNoErr )
	{
		to.APND(" ORDER BY RAND()");
	}
}

void local_addOperator( STRING_TYPE & to, LP_TypeDesc op, const char * value, unsigned int size )
{
	switch( op )
	{
	case kLassoEquals:
		if ( value == NULL )
			to.APND(" IS ");
		else
			to.APND(" = ");
		break;
	case kLassoGreaterThan:
		to.APND(" > ");
		break;
	case kLassoGreaterThanEquals:
		to.APND(" >= ");
		break;
	case kLassoLessThan:
		to.APND(" < ");
		break;
	case kLassoLessThanEquals:
		to.APND(" <= ");
		break;
	case kLassoNOT:
		if ( value == NULL )
			to.APND(" IS NOT ");
		else
			to.APND(" <> ");
		break;
	case kLassoContains:
	case kLassoBeginsWith:
	case kLassoEndsWith:
		to.APND(" LIKE ");
		break;
	case kLassoNotContains:
	case kLassoNotBeginsWith:
	case kLassoNotEndsWith:
		to.APND(" NOT LIKE ");
		break;
	case kLassoInFullText:
		to.APND(" AGAINST ");
		break;
	case kLassoInRegExp:
		to.APND(" REGEXP ");
		break;
	case kLassoNotInRegExp:
		to.APND(" NOT REGEXP ");
		break;
	case kLassoInList:
		to.APND(" IN ");
		break;
	case kLassoNotInList:
		to.APND(" NOT IN ");
		break;
	default:
		to.APND(" = ");
		break;
	}
}

void local_addSearchValue( STRING_TYPE & to, const char * value, unsigned int size, LP_TypeDesc op )
{
	if ( value == NULL )
	{
		to.APND("NULL");
	}
	else if ( value[0] == '\0' )
	{
		to.APND("''");
	}
	else
	{
		// values need to be encoded
		char * encoded = new char[size*2+1];

		// mysql_real_escape_string didn't seem to be exported in the mysqlk client I am using
		 mysql_escape_string(encoded, value, size);	
		
		if ( op == kLassoEndsWith || op == kLassoContains || op == kLassoNotEndsWith || op == kLassoNotContains || 
				  op == kLassoBeginsWith || op == kLassoContains || op == kLassoNotBeginsWith || op == kLassoNotContains )
		{
			to.APND("'");
			if (op == kLassoEndsWith || op == kLassoContains || op == kLassoNotEndsWith || op == kLassoNotContains)
				to.APND("%");
			to.APND(encoded);
			if (op == kLassoBeginsWith || op == kLassoContains || op == kLassoNotBeginsWith || op == kLassoNotContains)
				to.APND("%");
			to.APND("'");
		}
		else
		{	
			if ( op == kLassoInFullText )
				to.APND("(");
			to.APND("'").APND(encoded).APND("'");
			if ( op == kLassoInFullText )
				to.APND(")");
		}

		delete [] encoded;
	}
}
