<?LassoScript

// [Log] Redefining log tag with encoding safe replacement.

$__tags__->(insert: '_log' = \log);
$__tags__->(remove: 'log');

define_tag: 'log', -optional='file', -optional='encoding', -privileged, -container;
	if: !(local_defined: 'encoding');
		if: ((lasso_version: -lassoplatform) >> 'mac');
			local: 'encoding' = 'mac';
		else;
			local: 'encoding' = 'iso-8859-1';
		/if;
	/if;
	if: (local_defined: 'file');
		local: '_file' = (string: #file);
		#_file->trim;
		local: '_file' = (string_replace: #_file, -find='\\', -replace='/');
		fail_if: #_file->(beginswith: '///'), -9951, '[Log] Cannot write to fully qualified paths';
		fail_if: #_file->(endswith: '/'), -9951, '[Log] Cannot write to a directory';
		fail_if: (#_file >> '//'), -9951, '[Log] Cannot write to fully qualified paths';
		fail_if: (#_file >> '..'), -9951, '[Log] Cannot use .. in log paths';
		if: !(file_exists: #_file);
			file_create: #_file;
			fail_if: (file_currenterror: -errorcode) != 0, (file_currenterror: -errorcode), '[Log] ' + file_currenterror;
		/if;
		local: 'bytes' = bytes;
		#bytes->(importstring: run_children, #encoding);
		file_write: #file, #bytes;
	else;
		log_critical: (string: run_children);
	/if;
/define_tag;


// [Database_FMContainer] tag that fetches the data from an FM container field

define_tag: 'Database_FMContainer', 
		-required='field', 
		-optional='type', 
		-privileged,
		-priority='replace';

	fail_if: (database_name == ''), -1, 'Error: Database not found or enabled.';
	fail_if: (table_name == ''), -1, 'Error: Table not found or enabled.';
	fail_if: ((field: #field) == null), -1, 'Error: Field not found or enabled.';

	local: 'table' = (table_realname: database_name, table_name);
	local: 'keyvalue' = (keyfield_value);

	inline:
		-database = 'lasso_internal',
		-sql = '\
			select \
				a.name,
				b.name as host, \
				b.port, \
				case \
					when a.always_use_defaults = \'Y\' then \
						b.username \
					else \
						a.username \
				end as username, \
				case \
					when a.always_use_defaults = \'Y\' then \
						b.password \
					else \
						a.password \
				end as password, \
				b.schema, \
				c.name as connector \
			from \
				security_datasource_databases as a \
				left join security_datasource_hosts as b on (a.id_host = b.id) \
				left join security_datasources as c on (b.id_datasource = c.id) \
			where \
				case \
					when a.alias is not null and a.alias <> \'\' then \
						a.alias \
					when a.name is not null and a.name <> \'\' then \
						a.name \
				end = \'' (encode_sql: database_name ) '\' and \
				a.enabled = \'Y\' and \
				b.enabled = \'Y\' and \
				c.name in (\'FMServerAdvancedDataSource\', \'FMRemoteDataSource\') \
			limit 1';

		fail_if: !found_count, -1728, 'The database "' database_name '" was not found, was not enabled, or was not a FileMaker Pro or FileMaker Server Advanced database.';

		local: 'database' = (field: 'name');
		local: 'username' = (field: 'username');
		local: 'password' = (field: 'password');
		local: 'url' = (field: 'host');
		local: 'port' = (field: 'port');
		local: 'connector' = (field: 'connector');

		if: !#username->size;
			#username = 'dummy';
		/if;
	/inline;

	local: 'result' = null;
	if: #connector == 'FMRemoteDataSource';
		local: '_type' = (string: (local: 'type'));
		if: (#_type == '');
			local: '_type' = 'jpeg';
		/if;
		local: 'url' = 'http://' + #url + ':' + #port + '/FMPro?-db=' + #database + '&-lay=' + #table + '&' + #field + '=jpeg&-recid=' + #keyvalue + '&-img';
		local: 'result' = (include_url: #url, -username = #username, -password = #password);
	else: #connector == 'FMServerAdvancedDataSource';
		fail_if: ((field: #field) !>> '/fmi/xml/cnt/data'), -1, 'Error: Field is not a container field.';
		local: 'url' = 'http://' + #url + ':' + #port + (field: #field);
		local: 'result' = (include_url: #url, -username = #username, -password = #password);
	/if;
	
	if: (#result >> 'HTTP 404: Not Found');
		return: null;
	/if;
	
	return: @#result;

/define_tag;



// replacement for old repetition functionality
define_type:'repetition';

	//
	// IMPORTANT - [Repetition] has been deprecated.
	//
	// [If: (Repetition == 2)] should be changed to [If: (Loop_Count % 2 == 0)]
	//

    define_tag: 'onCompare', -required='modulo';
        
        if: (var_defined: '__loop_count__') && ($__loop_count__->size > 0);
        	return: (((integer: $__loop_count__->last) % #modulo) != 0);
        /if;
        
        if: !(var_defined: '__repetition__');
        	var: '__repetition__' = (map);
        	log_warning: 'Warning: The [Repetition] tag has been deprecated.';
        /if;
        
        local: 'temp' = ((integer: $__repetition__->(find: #modulo)) + 1);
        $__repetition__->(insert: #modulo = #temp);
        return: ((#temp % #modulo) != 0);
        
    /define_tag;
    
/define_type;

// replacement for the old client_ip == '127.0.0.*' functionality
define_type:'client_ip';
	local: 'data' = _client_ip_old;

	// do smart comparisons
	define_tag:'onCompare';

		// params may be a list item comma delimited items
		// these are OR not AND

		local: 'mainList' = (string: params->(get: 1))->(split: ',');

		if: #mainList->size == 0; // no items
			return: -1;
		/if;

		local: 'lhs' = self->data->(split: '.');

		iterate: #mainList, (local: 'i');
			#i->trim;

			if: (string_isdigit: #i);
				#i = (_ip_frominteger: #i);
			/if;

			if: #i->size == 0;
				return: -1;
			/if;

			local: 'rhs' = #i->(split: '.');
			local: 'result' = 0;

			iterate: #rhs, (local: 'j');
				if: (#lhs->(get: (loop_count)) != #j) && (#j != '*');
					#result = -1;
				/if;
			/iterate;
			return: #result;
		/iterate;

		return: -1; // just in case
	/define_tag;

	// when called normally, need to output the client_ip_old tag
	define_tag:'onConvert';
		return: self->data;
	/define_tag;
/define_type;

// turn an ip address into an integer
define_tag: '_ip_tointeger',
	-required = 'ip',
	-type = 'string';

	fail_if: #ip->(split: '.')->size != 4, -1, 'Incorrect quad decimal IP address';

	local:
		'result' = 0,
		'm' = 256 * 256 * 256;

	iterate: #ip->(split: '.'), (local: 'i');
		#result += (integer: #i) * #m;
		#m /= 256;
	/iterate;

	return: #result;
/define_tag;

// turn an integer into an IP address
define_tag: '_ip_frominteger',
	-required = 'ip';

	local:
		'x' = (integer: #ip),
		'result' = array,
		'm' = 256 * 256 * 256;

	loop: 4;
		#result->(insert: #x / #m);
		#x %= #m;
		#m /= 256;
	/loop;

	return: (string: #result->(join: '.'));
/define_tag;

// replacement for the old client_address == '*.edu' functionality
define_type: 'client_address';
	local: 'data' = _client_address_old;

	// do smart comparisons
	// works almost the same as client_ip, except compares from back to front
	define_tag: 'onCompare';

		// simple check to see if they do equal each other first
		if: (string: params->(get: 1)) == self->data;
			return: 0;
		/if;

		// params may be a comma dlist item comma delimited items
		local:'mainList' = (string: params->(get: 1))->(split:',');

		if: #mainList->size == 0; // no items
			return: -1;
		/if;

		iterate: #mainList, (local: 'i');
			#i->trim;

			#i->(replace: '.', '\\.');
			#i->(replace: '*', '.*');
			protect;
				if: (string_findregexp: -find = #i, self->data)->size == 1;
					return: 0;
				/if;
				handle_error;
					return: -1;
				/handle_error;
			/protect;
		/iterate;

		return: -1;
	/define_tag;

	// when called normally, need to output the client_ip_old tag
	define_tag:'onConvert';
		return: self->'data';
	/define_tag;

/define_type;

define_type:'client_addr', 'client_address';
/define_type;

?>