<!--	
	Database Check
	
	This script performs a routine check on all Lasso MySQL databases at startup.
	It runs before LassoStartup.Lasso to ensure that any databases which were not
	closed properly have been repaired if necessary.
	    
-->

<?LassoScript
	{
		log_warning: 'Checking/Repairing Lasso MySQL Databases';
		
		inline: -database='lasso_internal', -sql='SHOW DATABASES';
			records;
				local: 'database' = '`' + (encode_sql: (field: (field_name: 1))) + '`';
				local: 'tables' = (array);
				local: 'repair' = (array);
				inline: -sql=('SHOW TABLES FROM ' + #database);
					records;
						#tables->(insert: #database + '.`' + (encode_sql: (field: (field_name: 1))) + '`');
					/records;
				/inline;

				if: (#tables->size > 0);
					log_warning: '    Checking: ' + #database;
					iterate: #tables, (local: 'table');
						inline: -sql=('CHECK TABLE ' + #table + ' FAST');
							records;
								if: ((field: 'msg_type') !>> 'status') && ((field: 'msg_text') !>> 'closed the table properly');
									log_warning: '        ' + (field: 'table') + ': ' + (field: 'msg_type') + ' ' + (field: 'msg_text');									
								/if;
								if: (#repair >> #table) || ((field: 'msg_type') !>> 'status');
									if: ((field: 'msg_type') >> 'status');
										#repair->(removeall: #table);
									else: !(#repair >> #table);
										#repair->(insert: #table);
									/if;
								/if;
							/records;
						/inline;
					/iterate;
				/if;

				if: (#repair->size > 0);
					Log_Critical: '    Repairing: ' + #database;
					
					iterate: #repair, (local: 'table');
						inline: -sql=('REPAIR TABLE ' + #table);
							records;
								Log_Critical: '        ' + (field: 'table') + ': ' + (field: 'msg_type') + ' ' + (field: 'msg_text');									
							/records;
						/inline;
					/iterate;
				/if;
				
			/records;
		/inline;
		
		//
		// Errors - Removes any errors in excess of 10,000 in the _errors table
		//
		
		// Fix encoding on errors table
	
		inline: -search,
				-database='lasso_internal',
				-table='security_datasource_databases',
				-keyfield='id',
				-op='eq', 'id_datasource'='1',
				-op='eq', 'id_host'='1',
				-op='eq', 'name'='lasso_internal';
			if: found_count == 1;
				inline: -search,
						-database='lasso_internal',
						-table='security_database_tables',
						-keyfield='id',
						-op='eq', 'id_database'=(field: 'id'),
						-op='eq', 'name'='_errors',
						-op='neq', 'encoding'='utf-8';
					if: found_count == 1;
						inline: -update,
								-database='lasso_internal',
								-table='security_database_tables',
								-keyfield='id',
								-keyvalue=(field: 'id'),
								'encoding'='utf-8';
						/inline;
					/if;
				/inline;
			/if;
		/inline;

		// Delete excess lines
		
		inline: -database='lasso_internal', -sql='SELECT COUNT(*) FROM _errors';
			local: 'error_count' = (integer: (field: (field_name:1)));
			
			if: #error_count > 1000;
			
				log_warning: '    Errors Table: Deleting excess error messages.';
			
				inline: -database='lasso_internal', -sql='DELETE FROM _errors ORDER BY id LIMIT ' + (#error_count - 1000);
				/inline;
			
				inline: -database='lasso_internal', -sql='OPTIMIZE TABLE _errors';
				/inline;
			
			/if;
			
		/inline;
						
		//
		// LDML Reference - Creates the comment table if it doesn't exist!
		//
		
		protect;
			inline: -database='LDML7_Reference', -sql='CREATE TABLE IF NOT EXISTS `Comments` ( 
				`id` MEDIUMINT(9) NOT NULL auto_increment, 
				`tag_id` MEDIUMINT(9) NOT NULL DEFAULT "0", 
				`tag_name` VARCHAR(64), 
				`com_date` VARCHAR(32), 
				`com_author` VARCHAR(255) NOT NULL DEFAULT "", 
				`com_email` VARCHAR(255), 
				`com_subject` VARCHAR(255), 
				`com_comments` TEXT, 
				`com_stamp` TIMESTAMP NOT NULL, 
				PRIMARY KEY (`id`), 
				KEY (`tag_id`), 
				KEY (`tag_name`), 
				KEY (`com_author`), 
				KEY (`com_subject`), 
				KEY (`com_stamp`) 
			)'; 
			/inline;
		/protect;
		
		log_warning: 'Checking/Repairing Done';

	}->(run);
?>
