<?LassoScript
	
	// [XML_Extract: -XML='xml data', -XPath='xpath']
	// [XML_Extract: -File='path to xml file', -XPath='xpath']
	//
	// Other parameter names are allowed since the documentation for LP6 was unclear.

	Define_Tag: 'XML_Extract',
			-Optional='XML',
			-Optional='XPath',
			-Optional='XMLData',
			-Optional='File',
			-Optional='XMLFile',
			-Priority='replace';
		
		if: (local_defined: 'xml');
			if: (#xml->isa: 'xml');
				local: 'data' = #xml;
			else;
				local: 'data' = (xml: #xml);
			/if;
		else: (local_defined: 'xmldata');
			if: (#xmldata->isa: 'xml');
				local: 'data' = #xmldata;
			else;
				local: 'data' = (xml: #xmldata);
			/if;
		else: (local_defined: 'file');
			local: 'data' = (xml: (include_raw: #file));
		else: (local_defined: 'xmlfile');
			local: 'data' = (xml: (include_raw: #xmlfile));
		else;
			fail: (-1), '[XML_Extract] No -XML or -File parameter was specified.';
		/if;
			
		if: (local_defined: 'xpath');
			local: 'path' = (string: #xpath);
		else;
			fail: (-1), '[XML_Extract] No -XPath parameter was specified.';
		/if;

		return: #data->(extract: #path);
		
	/Define_Tag;
	
	// [XML_Transform: -XML='xml data', -XSL='xpath']
	// [XML_Transform: -File='path to xml file', -XSL='style sheet']
	// [XML_Transform: -XML='xml data', -XSLFile='path to style sheet']
	//
	// Other parameter names are allowed since the documentation for LP6 was unclear.

	Define_Tag: 'XML_Transform',
			-Optional='XML',
			-Optional='XSL',
			-Optional='XMLData',
			-Optional='File',
			-Optional='XMLFile',
			-Optional='XSLT',
			-Optional='XSLFile',
			-Optional='XSLTFile',
			-Priority='replace';
		
		if: (local_defined: 'xml');
			if: (#xml->isa: 'xml');
				local: 'data' = #xml;
			else;
				local: 'data' = (xml: #xml);
			/if;
		else: (local_defined: 'xmldata');
			if: (#xmldata->isa: 'xml');
				local: 'data' = #xmldata;
			else;
				local: 'data' = (xml: #xmldata);
			/if;
		else: (local_defined: 'file');
			local: 'data' = (xml: (include_raw: #file));
		else: (local_defined: 'xmlfile');
			local: 'data' = (xml: (include_raw: #xmlfile));
		else;
			fail: (-1), '[XML_Transform] No -XML or -File parameter was specified.';
		/if;
			
		if: (local_defined: 'xsl');
			local: 'style' = (string: #xsl);
		else: (local_defined: 'xslt');
			local: 'style' = (string: #xslt);
		else: (local_defined: 'xslfile');
			local: 'style' = (include_raw: #xslfile);
		else: (local_defined: 'xsltfile');
			local: 'style' = (include_raw: #xsltfile);
		else;
			fail: (-1), '[XML_Transform] No -XSL parameter was specified.';
		/if;

		return: #data->(transform: #style);
		
	/Define_Tag;


	//
	// XML Read/Write
	//
	// This code defines the [XML_RW] type.  This type can be used by developers who need
	// similar read/write capabilities to the XML tags from LP6.
	//
	// This type supports the following tags:
	//
	// [XML_RW: 'xml'] accepts XML data and instantiates a new type.
	// [XML_RW->Children] (read/write) returns an array of children.
	// [XML_RW->Contents] (read/write) returns the rendered contents of the root tag
	// [XML_RW->Attributes] (read/write) returns an array of pairs.
	// [XML_RW->Name] (read/write) returns the name of the root tag.
	// [XML_RW->Transform] requires an XSLT style sheet, returns the results.
	// [XML_RW->Extract] requires an XPath, returns the resulting array.
	// [XML_RW->ExtractOne] requires an XPath, returns the first element of the resulting array.
	//
	// When cast to string the XML is reassembled with any new modifications.
	//
	
	define_type: 'xml_rw';
	
		local: 'children' = (array);
		local: 'attributes' = (array);
		local: 'contents' = '';
		local: 'name' = '';
		local: 'refresh' = false;
		
		define_tag: 'oncreate', -required='xml';
			if: #xml->(isa: 'xml');
				local: 'temp' = #xml;
			else;
				local: 'temp' = (xml: (string: #xml));
			/if;
			self->'children' = #temp->children;
			self->'attributes' = #temp->attributes;
			self->'contents' = #temp->contents;
			self->'name' = #temp->name;
			self->'refresh' = false;
		/define_tag;
		
		define_tag: 'onconvert', -required='type';
			local: 'output' = '';
			#output += '<' + self->'name';
			#output += self->(_params: self->'attributes');
			if: (self->'refresh' == true);
				if: self->'contents'->size > 0;
					#output += '>';
					#output +=  self->'contents';
					#output += '</' + self->'name' + '>';
				else;
					#output += ' />';
				/if;
			else;
				if: self->'children'->size > 0;
					#output += '>';
					#output +=  self->'children'->(join: '');
					#output += '</' + self->'name' + '>';
				else;
					#output += ' />';
				/if;
			/if;
			return: @#output;
		/define_tag;
		
		define_tag: 'children';
			self->_refresh;
			iterate: self->'children', (local: 'value');
				local: 'string' = (string: #value);
				if: #string->(beginswith: '<');
					#value = (xml_rw: #value);
				else;
					#value = #string;
				/if;
			/iterate;
			return: self->'children';
		/define_tag;
		
		define_tag: 'contents';
			self->_refresh;
			self->'contents' = self->'children'->(join: '');
			self->'refresh' = true;
			return: @(self->'contents');
		/define_tag;
	
		define_tag: 'extract', -required='xpath';
			self->refresh;
			return: (xml_extract: self, -xpath=#xpath);
		/define_tag;
		
		define_tag: 'extractone', -required='xpath';
			self->_refresh;
			local: 'temp' = (xml_extract: self, -xpath=#xpath);
			if: #temp->size > 0;
				return: #temp->first;
			/if;
		/define_tag;
	
		define_tag: 'transform', -required='xslt';
			self->_refresh;
			return: (xml_transform: self, -xslt=#xslt);
		/define_tag;
		
		define_tag: '_refresh';
			if: (self->'refresh' == true);
				self->(oncreate: (string: self));
			/if;
			self->'refresh' = false;
		/define_tag;
	
		define_tag: '_params', -required='params';
			local: 'output' = '';
			iterate: #params, (local: 'temp');
				if: #temp->(isa: 'pair');
					#output += ' ' + #temp->name + '="' + (encode_xml: #temp->value) + '"';
				else: #temp->(isa: 'array') || #temp->(isa: 'map');
					#output += self->(_params: #temp);
				/if;
			/iterate;
			return: @#output;
		/define_tag;
	
	/define_type;

?>
